import boto3
import google.auth.transport.requests
from google.oauth2 import id_token
import requests

METADATA_HEADERS = {"Metadata-Flavor": "Google"}
METADATA_VM_IDENTITY_URL = (
    "http://metadata.google.internal/computeMetadata/v1/"
    "instance/service-accounts/default/identity?"
    "audience={audience}&format={format}&licenses={licenses}"
)
FORMAT = "full"
LICENSES = "TRUE"

class GcpAwsAuth:
    def __init__(self, iam_role_arn, audience, aws_region):
        self.iam_role_arn = iam_role_arn
        self.audience = audience
        self.aws_region = aws_region

    def acquire_token(self, format: str = "standard", licenses: bool = True
    ) -> str:
        url = METADATA_VM_IDENTITY_URL.format(
            audience=self.audience, format=format, licenses=licenses
        )
        r = requests.get(url, headers=METADATA_HEADERS)
        r.raise_for_status()
        return r.text

    def verify_token(self, token: str) -> dict:
        request = google.auth.transport.requests.Request()
        payload = id_token.verify_token(token, request=request, audience=self.audience)
        return payload

    def get_boto_session(self):
        token = self.acquire_token()
        self.verify_token(token)
        sts_client = boto3.client("sts", region_name=self.aws_region)

        cred = sts_client.assume_role_with_web_identity(
            RoleArn=self.iam_role_arn,
            RoleSessionName='aws_session',
            WebIdentityToken=token,
            DurationSeconds=900
        )
        return boto3.Session(
            aws_access_key_id=cred.get('Credentials').get("AccessKeyId"),
            aws_secret_access_key=cred.get('Credentials').get("SecretAccessKey"),
            aws_session_token=cred.get('Credentials').get("SessionToken"),
            region_name=self.aws_region
        )
