import gzip
import json
import os
import subprocess
import uuid

from io import BytesIO
from dynamodb_json import json_util
from flask import Flask

import log_util
from gcp_aws_auth import GcpAwsAuth
from s3_client import S3Client
from sqs import SQSQueue


class SQSMessageProcessor(SQSQueue):
    def __init__(self):
        self.app = Flask(__name__)
        self.logger = log_util.get_logger("mongo_migrator", "INFO")
        self.load_config()
        self.gcp_aws_auth = GcpAwsAuth(
            iam_role_arn=self.web_identity_iam_role_arn,
            audience=self.web_identity_audience,
            aws_region=self.aws_region
        )
        SQSQueue.__init__(
            self,
            queue_name=self.queue_name,
            queue_region=self.aws_region,
            logger=self.logger,
            num_messages=1,
            sqs_endpoint_url=self.endpoint_url,
            web_identity_iam_role_arn=self.web_identity_iam_role_arn,
            web_identity_audience=self.web_identity_audience,
        )
        self.logger.info("...SQS listener Initialized")
        self._set_up_clients()

    def load_config(self):
        self.aws_region = os.getenv("AWS_REGION", "us-east-1")
        self.deployment = os.getenv("DEPLOYMENT", "AWS")
        self.queue_name = os.getenv("QUEUE_NAME", "temp-migration-queue")
        self.endpoint_url = os.getenv("AWS_ENDPOINT_URL")

        # MONGO DB properties
        self.mongo_database = os.getenv("MONGO_DATABASE", "default")
        self.mongo_table = os.getenv("MONGO_table", "test")
        self.mongo_host = f'{os.getenv("MONGO_HOST", "localhost")}:{os.getenv("MONGO_PORT", "27017")}'
        self.mongo_user = os.getenv("MONGO_USER")
        self.mongo_password = os.getenv("MONGO_PASSWORD")
        self.mongo_id_fields = (
            os.getenv("MONGO_ID_FIELDS").strip().split(",")
            if os.getenv("MONGO_ID_FIELDS") else None
        )
        self.insert_chunk_size = os.getenv("INSERT_CHUNK_SIZE", 100000)

        # Web Identity properties
        self.web_identity_iam_role_arn = os.getenv("WEB_IDENTITY_IAM_ROLE_ARN")
        self.web_identity_audience = os.getenv("WEB_IDENTITY_AUDIENCE")

    def _set_up_clients(self):
        self.s3_client = S3Client(self.boto_session)

    def process_message(self, message):
        self.logger.info(f"Processing message: {message}")
        data = self.get_json_data_from_s3(message)
        for record in self.chunk(data, self.insert_chunk_size):
            self.mongo_insert(record)
        self.sqs_client.delete_message(self.queue_url, message["ReceiptHandle"])

    def get_json_data_from_s3(self, message):
        parsed_message = json.loads(message["Body"])
        s3 = parsed_message.get("Records")[0].get("s3")
        bucket = s3.get("bucket").get("name")
        obj_key = s3.get("object").get("key")
        self.logger.info("Processing {} from bucket {}".format(obj_key, bucket))
        with BytesIO() as zf:
            self.get_s3_client().download_fileobj(
                file_obj=zf, bucket_name=bucket, key_name=obj_key
            )
            file = zf.read()
        data = gzip.decompress(file).decode("utf-8").split("\n")
        return [self.translate_to_json(line) for line in data]

    def translate_to_json(self, line):
        dictionary = json_util.loads(json.loads(line).get("Item"))
        # Construct mongo DB ID field
        if self.mongo_id_fields:
            dictionary['_id'] = "_".join(
                dictionary.get(field.strip())
                for field in self.mongo_id_fields
                if field and field.strip()
            )
        return json.dumps(dictionary)

    def mongo_insert(self, data):
        self.logger.info("Inserting {} records".format(len(data)))
        filename = "/tmp/{}.json".format(str(uuid.uuid4()))
        with open(filename, 'w') as f:
            f.write("\n".join(data))
        cmd = [
            'mongoimport',
            '--db', self.mongo_database,
            '--collection', self.mongo_table,
            '--host', self.mongo_host,
            '--type', 'json',
            '--file', filename,
            "--mode", "upsert"
        ]
        if self.mongo_user and self.mongo_password:
            cmd.extend([
                '--user', self.mongo_user,
                '--password', self.mongo_password
            ])
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        for line in p.stdout:
            self.logger.info(line)
        p.wait()
        self.logger.info("Insert command status {}".format(p.returncode))
        os.remove(filename)
        if not p or p.returncode != 0:
            raise Exception("Mongo import failed for file {}.".format(filename))

    def get_s3_client(self):
        if self.deployment == "AWS":
            return self.s3_client
        elif self.deployment == "GCP":
            return S3Client(self.gcp_aws_auth.get_boto_session())

    @staticmethod
    def chunk(iterable, chunk_size):
        items = []
        for item in iterable:
            items.append(item)
            if len(items) != chunk_size:
                continue
            yield items
            items = []
        if len(items) > 0:
            yield items


def start_service_mcs2(port):
    service = SQSMessageProcessor()
    return service.app
