import threading
from time import sleep
from boto3 import Session

from gcp_aws_auth import GcpAwsAuth


class SqsClient(object):
    def __init__(self, boto_session,
                 boto_config=None,
                 endpoint_url=None,
                 deployment=None,
                 web_identity_iam_role_arn=None,
                 web_identity_audience=None,
                 aws_region=None):
        self.sqs_client = boto_session.client(
            "sqs", config=boto_config, endpoint_url=endpoint_url
        )
        self.deployment = deployment
        self.boto_config = boto_config
        self.endpoint_url = endpoint_url
        self.gcp_aws_auth = GcpAwsAuth(
            iam_role_arn=web_identity_iam_role_arn,
            audience=web_identity_audience,
            aws_region=aws_region
        )

    def get_sqs_client(self):
        if self.deployment == "AWS":
            return self.sqs_client
        elif self.deployment == "GCP":
            boto_session = self.gcp_aws_auth.get_boto_session()
            return boto_session.client(
                "sqs", config=self.boto_config, endpoint_url=self.endpoint_url
            )

    def get_queue_url(self, queue_name):
        return self.sqs_client.get_queue_url(QueueName=queue_name)["QueueUrl"]

    def receive_message(
            self, queue, wait_seconds=None, num_messages=1, attributes=None
    ):
        if attributes is None:
            attributes = list()
        return self.sqs_client.receive_message(
            QueueUrl=queue,
            WaitTimeSeconds=wait_seconds,
            MaxNumberOfMessages=num_messages,
            MessageAttributeNames=attributes,
        )

    def delete_message(self, queue, receipt_handle):
        return self.sqs_client.delete_message(
            QueueUrl=queue, ReceiptHandle=receipt_handle
        )

    def change_message_visibility(self, queue, receipt_handle, timeout):
        return self.sqs_client.change_message_visibility(
            QueueUrl=queue, ReceiptHandle=receipt_handle, VisibilityTimeout=timeout
        )


class SQSQueue(object):
    def __init__(
            self,
            queue_name,
            queue_region,
            logger,
            num_messages=1,
            attributes=None,
            access_key_id=None,
            access_key=None,
            sqs_endpoint_url=None,
            deployment="GCP",
            web_identity_iam_role_arn=None,
            web_identity_audience=None
    ):
        self.logger = logger
        self.boto_session = Session(region_name=queue_region)
        if access_key_id and access_key:
            self.process_thread_boto_session = Session(
                aws_access_key_id=access_key_id,
                aws_secret_access_key=access_key,
                region_name=queue_region,
            )
        else:
            self.process_thread_boto_session = Session(region_name=queue_region)

        self.sqs_client = SqsClient(
            boto_session=self.boto_session, endpoint_url=sqs_endpoint_url,
            deployment=deployment,
            web_identity_iam_role_arn=web_identity_iam_role_arn,
            web_identity_audience=web_identity_audience,
            aws_region=queue_region
        )

        self.queue_url = self.sqs_client.get_queue_url(queue_name)
        self.queue_region = queue_region
        self.listen_enabled = True
        self.num_messages = num_messages
        self.attributes = attributes
        self.first_boot = True

        self.logger_extra = {"message_id": None}

        self._start_sqs_loop()

    def post(self):
        if self.listen_enabled:
            return 200
        self.listen_enabled = True
        return 200

    def delete(self):
        self.listen_enabled = False
        self.sqs_loop_thread.join()
        self.logger.debug("SQS Listen stopped")
        return 200

    def get(self):
        return {"ListenStatus": str(self.listen_enabled)}, 200

    def _start_sqs_loop(self):
        self.logger.debug("Creating SQS Listen thread")
        self.sqs_loop_thread = threading.Thread(target=self.sqs_loop)
        self.sqs_loop_thread.daemon = True
        self.sqs_loop_thread.start()

    def sqs_loop(self):
        while True:
            try:
                self.logger.debug("SQS Listen thread started")

                while self.listen_enabled:
                    # Sleep is enabled - don't read any messages from the queue.
                    if self.check_sleep():
                        continue
                    messages = self.sqs_client.receive_message(
                        queue=self.queue_url,
                        num_messages=self.num_messages,
                        wait_seconds=20,
                        attributes=self.attributes,
                    )

                    for message in messages.get("Messages", list()):
                        self.process_message(message)

                self.logger.warning("Exiting SQS Listen thread")
            except Exception:
                self.logger.exception(
                    "sqs_loop: exception encountered, waiting and retrying sqs connection"
                )
                sleep(10)

    def check_sleep(self):
        if self.first_boot:
            sleep(5)
            self.first_boot = False

    def process_message(self, message):
        pass
